from __future__ import annotations
from typing import TypeVar, TYPE_CHECKING

import os
import re
import time
import warnings
from copy import deepcopy

import numpy as np
import scipy.sparse as sparse
from ruamel.yaml import YAML, add_constructor, SafeConstructor

if TYPE_CHECKING:  # pragma: no cover
    from ._experiment import Experiment
    from ._solutions import BaseSolution, StepSolution, CycleSolution

    Solution = TypeVar('Solution', bound='BaseSolution')


class Model:
    """Circuit model."""

    def __init__(self, params: dict | str = 'params.yaml'):
        """
        A class to construct and run the model. Provide the parameters using
        either a dictionary or a '.yaml' file. Note that the number of Rj and
        Cj attributes must be consistent with the num_RC_pairs value. See the
        notes for more information on the callable parameters.

        Parameters
        ----------
        params : dict | str
            Mapping of model parameter names to their values. Can be either
            a dict or absolute/relateive file path to a yaml file (str). The
            keys/value pair descriptions are given below. The default uses an
            internal yaml file.

            ============= ========================== ================
            Key           Value                      *type*, units
            ============= ========================== ================
            num_RC_pairs  number of RC pairs         *int*, -
            soc0          initial state of charge    *float*, -
            capacity      maximum battery capacity   *float*, Ah
            ce            coulombic efficiency       *float*, -
            mass          total battery mass         *float*, kg
            isothermal    flag for isothermal model  *bool*, -
            Cp            specific heat capacity     *float*, J/kg/K
            T_inf         room/air temperature       *float*, K
            h_therm       convective coefficient     *float*, W/m2/K
            A_therm       heat loss area             *float*, m2
            ocv           open circuit voltage       *callable*, V
            R0            series resistance          *callable*, Ohm
            Rj            resistance in RCj          *callable*, Ohm
            Cj            capacity in RCj            *callable*, F
            ============= ========================== ================

        Raises
        ------
        TypeError
            'params' must be type dict or str.
        ValueError
            'params' contains invalid and/or excess key/value pairs.

        Warning
        -------
        A pre-processor runs at the end of the model initialization. If you
        modify any parameters after class instantiation, you will need to
        re-run the pre-processor (i.e., the ``pre()`` method) afterward.

        Notes
        -----
        The 'ocv' property needs a signature like ``f(soc: float) -> float``,
        where 'soc' is the state of charge. All R0, Rj, and Cj properties need
        signatures like ``f(soc: float, T_cell: float) -> float``. 'T_cell' is
        the cell temperature in K.

        Rj and Cj are not real property names. These are used generally in the
        documentation. If ``num_RC_pairs=1`` then in addition to R0, you should
        define R1 and C1. If ``num_RC_pairs=2`` then you should also give R2
        and C2, etc. For the special case where ``num_RC_pairs=0``, you should
        not provide any resistance or capacitance values besides the series
        resistance R0, which is always required.

        """

        if isinstance(params, dict):
            params = params.copy()
        elif isinstance(params, str):
            params = _yaml_reader(params)

        self._repr_keys = [
            'num_RC_pairs',
            'soc0',
            'capacity',
            'ce',
            'mass',
            'isothermal',
            'Cp',
            'T_inf',
            'h_therm',
            'A_therm',
        ]

        self.num_RC_pairs = params.pop('num_RC_pairs')
        self.soc0 = params.pop('soc0')
        self.capacity = params.pop('capacity')
        self.ce = params.pop('ce')
        self.mass = params.pop('mass')
        self.isothermal = params.pop('isothermal')
        self.Cp = params.pop('Cp')
        self.T_inf = params.pop('T_inf')
        self.h_therm = params.pop('h_therm')
        self.A_therm = params.pop('A_therm')

        self.ocv = params.pop('ocv')
        self.R0 = params.pop('R0')

        for j in range(1, self.num_RC_pairs + 1):
            setattr(self, 'R' + str(j), params.pop('R' + str(j)))
            setattr(self, 'C' + str(j), params.pop('C' + str(j)))

        if len(params) != 0:
            extra_keys = list(params.keys())
            raise ValueError("'params' contains invalid and/or excess"
                             f" key/value pairs: {extra_keys=}")

        self.pre()

    def __repr__(self) -> str:  # pragma: no cover
        """
        Return a readable repr string.

        Returns
        -------
        readable : str
            A console-readable instance representation.

        """

        keys = self._repr_keys
        values = [getattr(self, k) for k in keys]

        summary = "\n    ".join([f"{k}={v}," for k, v in zip(keys, values)])

        readable = f"Model(\n    {summary}\n)"

        return readable

    def pre(self, initial_state: bool | Solution = True) -> None:
        """
        Pre-process and prepare the model for running experiments.

        This method builds solution pointers, registers algebraic variable
        indices, stores the mass matrix, and initializes the battery state.

        Parameters
        ----------
        initial_state : bool | Solution
            Control how the model state is initialized. If True (default), the
            state is set to a rested condition at 'soc0'. If False, the state
            is left untouched and only the parameters and pointers are updated.
            Given a Solution instance, the state is set to the final state of
            the solution. See notes for more information.

        Returns
        -------
        None.

        Warning
        -------
        This method runs during the class initialization. It generally does not
        have to be run again unless you modify model properties or attributes.
        You should manually re-run the pre-processor if you change properties
        after initialization. Forgetting to re-run the pre-processor can cause
        inconsistencies between the updated properties and the pointers, state,
        etc. If you are updating properties, but want the model's internal state
        to not be reset back to a rested condition, use the ``initial_state``
        option.

        Notes
        -----
        Using ``initial_state=False`` will raise an error if you are changing
        the size of your circuit (e.g., updating from one to two RC pairs).
        Without re-initializing, the model's state vector would be a different
        size than the circuit it is trying to solve. For this same reason, when
        initializing based on a Solution instance, the solution must also be
        the same size as the current model. In other words, a 1RC-pair model
        cannot be initialized given a solution from a 2RC-pair circuit.

        """

        from ._solutions import BaseSolution

        missing_attrs = []
        for j in range(1, self.num_RC_pairs + 1):
            if not hasattr(self, 'R' + str(j)):
                missing_attrs.append('R' + str(j))
            if not hasattr(self, 'C' + str(j)):
                missing_attrs.append('C' + str(j))

        if missing_attrs:
            raise AttributeError(f"Model is missing attrs {missing_attrs} to"
                                 " be consistent with 'num_RC_pairs'.")

        extra_attrs = []
        pattern = re.compile(r"^[RC](\d+)")
        for attr in list(self.__dict__.keys()):

            matches = pattern.match(attr)
            if matches is None:
                pass
            elif int(matches.group(1)) > self.num_RC_pairs:
                extra_attrs.append(attr)

        if extra_attrs:
            short_warn(f"Extra RC attributes {extra_attrs} are present, beyond"
                       " what was expected based on 'num_RC_pairs'.")

        ptr = {}
        ptr['soc'] = 0
        ptr['T_cell'] = 1
        ptr['eta_j'] = np.arange(2, 2 + self.num_RC_pairs)
        ptr['V_cell'] = self.num_RC_pairs + 2
        ptr['size'] = self.num_RC_pairs + 3

        algidx = [ptr['V_cell']]

        mass_matrix = np.zeros(ptr['size'])
        mass_matrix[ptr['soc']] = 1.
        mass_matrix[ptr['T_cell']] = self.mass*self.Cp*self.T_inf
        mass_matrix[ptr['eta_j']] = 1.

        sv0 = np.zeros(ptr['size'])
        sv0[ptr['soc']] = self.soc0
        sv0[ptr['T_cell']] = 1.
        sv0[ptr['eta_j']] = 0.
        sv0[ptr['V_cell']] = self.ocv(self.soc0)

        svdot0 = np.zeros_like(sv0)

        self._ptr = ptr
        self._algidx = algidx
        self._mass_matrix = sparse.diags(mass_matrix)

        self._t0 = 0.
        if isinstance(initial_state, BaseSolution):
            soln = deepcopy(initial_state)
            if soln.y[-1].size != sv0.size:
                raise ValueError("Cannot initialize state based on Solution"
                                 " object given in 'initial_state'. The model"
                                 " and solution have incompatible sizes.")

            self._sv0 = soln.y[-1].copy()
            self._svdot0 = soln.yp[-1].copy()

        elif not initial_state:
            if self._sv0.size != sv0.size:
                raise ValueError("The pre-processor failed. The model state is"
                                 " changing sizes but 'initial_state=False'.")

        elif initial_state:
            self._sv0 = sv0
            self._svdot0 = svdot0

    def rhs_funcs(self, t: float, sv: np.ndarray, inputs: dict) -> np.ndarray:
        """
        Right hand side functions.

        Returns the right hand side for the DAE system. For any differential
        variable i, rhs[i] must be equivalent to M[i, i]*y[i] where M is the
        mass matrix and y is an array of states. For algebraic variables rhs[i]
        must be an expression that equals zero.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        inputs : dict
            Dictionary detailing an experimental step.

        Returns
        -------
        rhs : 1D np.array
            The right hand side values of the DAE system.

        """

        rhs = np.zeros(self._ptr['size'])

        # state
        soc = sv[self._ptr['soc']]
        T_cell = sv[self._ptr['T_cell']]*self.T_inf
        eta_j = sv[self._ptr['eta_j']]
        voltage = sv[self._ptr['V_cell']]

        # state-dependent properties
        ocv = self.ocv(soc)
        R0 = self.R0(soc, T_cell)

        # calculated current and power
        current = -(voltage - ocv + np.sum(eta_j)) / R0
        power = current*voltage

        # state of charge (differential)
        ce = 1. if current >= 0. else self.ce
        rhs[self._ptr['soc']] = -ce*current / 3600. / self.capacity

        # temperature (differential)
        Q_gen = current*(ocv - voltage)
        Q_conv = self.h_therm*self.A_therm*(self.T_inf - T_cell)

        rhs[self._ptr['T_cell']] = (Q_gen + Q_conv) * (1 - self.isothermal)

        # RC overpotentials (differential)
        for j, ptr in enumerate(self._ptr['eta_j'], start=1):
            Rj = getattr(self, 'R' + str(j))(soc, T_cell)
            Cj = getattr(self, 'C' + str(j))(soc, T_cell)
            rhs[ptr] = -sv[ptr] / (Rj*Cj) + current / Cj

        # cell voltage (algebraic)
        mode = inputs['mode']
        units = inputs['units']
        value = inputs['value']

        if mode == 'current' and units == 'A':
            rhs[self._ptr['V_cell']] = current - value(t)
        elif mode == 'current' and units == 'C':
            rhs[self._ptr['V_cell']] = current - self.capacity*value(t)
        elif mode == 'voltage':
            rhs[self._ptr['V_cell']] = voltage - value(t)
        elif mode == 'power':
            rhs[self._ptr['V_cell']] = power - value(t)

        # values for rootfns
        total_time = self._t0 + t

        roots = {
            'soc': soc,
            'temperature_K': T_cell,
            'current_A': current,
            'current_C': current / self.capacity,
            'voltage_V': voltage,
            'power_W': power,
            'capacity_Ah': soc*self.capacity,
            'time_s': total_time,
            'time_min': total_time / 60.,
            'time_h': total_time / 3600.,
        }

        inputs['roots'] = roots

        return rhs

    def residuals(self, t: float, sv: np.ndarray, svdot: np.ndarray,
                  inputs: dict) -> np.ndarray:
        """
        Return the DAE residuals.

        The DAE residuals should be near zero at each time step. The solver
        requires the DAE to be written in terms of its residuals in order to
        minimize their values.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        svdot : 1D np.array
            State variable time derivatives at time t.
        inputs : dict
            Dictionary detailing an experimental step.

        Returns
        -------
        res : 1D np.array
            DAE residuals, res = M*yp - rhs(t, y).

        """
        return self._mass_matrix.dot(svdot) - self.rhs_funcs(t, sv, inputs)

    def run_step(self, exp: Experiment, stepidx: int) -> StepSolution:
        """
        Run a single experimental step.

        Parameters
        ----------
        exp : Experiment
            An experiment instance.
        stepidx : int
            Step index to run. The first step has index 0.

        Returns
        -------
        :class:`~thevenin.StepSolution`
            Solution to the experimental step.

        Warning
        -------
        The model's internal state is changed at the end of each experimental
        step. Consequently, you should not run steps out of order. You should
        always start with ``stepidx = 0`` and then progress to the subsequent
        steps afterward. Run ``pre()`` after your last step to reset the state
        back to a rested condition at 'soc0', if needed. Alternatively, you
        can continue running experiments back-to-back without a pre-processing
        in between if you want the following experiment to pick up from the
        same state that the last experiment ended.

        See also
        --------
        Experiment : Build an experiment.
        StepSolution : Wrapper for a single-step solution.

        Notes
        -----
        Using the ``run()`` loops through all steps in an experiment and then
        stitches their solutions together. Most of the time, this is more
        convenient. However, advantages for running step-by-step is that it
        makes it easier to fine tune solver options, and allows for analyses
        or control decisions in the middle of an experiment.

        """

        from ._ida_solver import IDASolver
        from ._solutions import StepSolution

        step = exp.steps[stepidx].copy()
        kwargs = exp._kwargs[stepidx].copy()

        if not callable(step['value']):
            value = step['value']
            step['value'] = lambda t: value

        kwargs['userdata'] = step
        kwargs['calc_initcond'] = 'yp0'
        kwargs['algebraic_idx'] = self._algidx

        if step['limits'] is not None:
            _setup_rootfn(step['limits'], kwargs)

        solver = IDASolver(self._residuals, **kwargs)

        start = time.time()
        ida_soln = solver.solve(step['tspan'], self._sv0, self._svdot0)
        timer = time.time() - start

        soln = StepSolution(self, ida_soln, timer)

        self._t0 = soln.t[-1]
        self._sv0 = soln.y[-1].copy()
        self._svdot0 = soln.yp[-1].copy()

        return soln

    def run(self, exp: Experiment, reset_state: bool = True,
            t_shift: float = 1e-3) -> CycleSolution:
        """
        Run a full experiment.

        Parameters
        ----------
        exp : Experiment
            An experiment instance.
        reset_state : bool
            If True (default), the internal state of the model will be reset
            back to a rested condition at 'soc0' at the end of all steps. When
            False, the state does not reset. Instead it will update to match
            the final state of the last experimental step.
        t_shift : float
            Time (in seconds) to shift step solutions by when stitching them
            together. If zero the end time of each step overlaps the starting
            time of its following step. The default is 1e-3.

        Returns
        -------
        :class:`~thevenin.CycleSolution`
            A stitched solution with all experimental steps.

        Warning
        -------
        The default behavior resets the model's internal state back to a rested
        condition at 'soc0' by calling the ``pre()`` method at the end of all
        steps. This means that if you run a second experiment afterward, it
        will not start where the previous one left off. Instead, it will start
        from the original rested condition that the model initialized with. You
        can bypass this by using ``reset_state=False``, which keeps the state
        at the end of the final experimental step.

        See also
        --------
        Experiment : Build an experiment.
        CycleSolution : Wrapper for an all-steps solution.

        """

        from ._solutions import CycleSolution

        solns = []
        for i in range(exp.num_steps):
            solns.append(self.run_step(exp, i))

        soln = CycleSolution(*solns, t_shift=t_shift)

        self._t0 = 0.
        if reset_state:
            self.pre()

        return soln

    def _residuals(self, t: float, sv: np.ndarray, svdot: np.ndarray,
                   res: np.ndarray, inputs: dict) -> None:
        """
        Solver-structured residuals.

        The IDASolver requires a residuals function in this exact form.
        Rather than outputting the residuals, the function returns None,
        but fills the 'res' input array with the DAE residuals.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        svdot : 1D np.array
            State variable time derivatives at time t.
        res : 1D np.array
            DAE residuals, res = M*yp - rhs(t, y).
        inputs : dict
            Dictionary detailing an experimental step.

        Returns
        -------
        None.

        """

        res[:] = self.residuals(t, sv, svdot, inputs)


class _RootFunction:
    """Root function callables."""

    def __init__(self, limits: tuple[str, float]) -> None:
        """
        This class is a generalized root function callable.

        Parameters
        ----------
        limits : tuple[str, float]
            A tuple of root function criteria arranged as ordered pairs of
            limit names and values, e.g., ('time_h', 10., 'voltage_V', 4.2).

        """

        self.keys = limits[0::2]
        self.values = limits[1::2]
        self.size = len(self.keys)

    def __call__(self, t: float, sv: np.ndarray, svdot: np.ndarray,
                 events: np.ndarray, inputs: dict) -> None:
        """
        Solver-structured root function.

        The IDASolver requires a root function in this exact form. Rather
        than outputting the events array, the function returns None, but
        fills the 'events' input array with root functions. If any 'events'
        index equals zero, the solver will exit prior to 'tspan[-1]'.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        svdot : 1D np.array
            State variable time derivatives at time t.
        events : 1D np.array
            An array of root/event functions. During integration, the solver
            will exit prior to 'tspan[-1]' if any 'events' index equals zero.
        inputs : dict
            Dictionary detailing an experimental step, with the 'roots' key
            added and filled within the `rhs_funcs()' method.

        Returns
        -------
        None.

        """

        for i, (key, value) in enumerate(zip(self.keys, self.values)):
            events[i] = inputs['roots'][key] - value


def _setup_rootfn(limits: tuple[str, float], kwargs: dict) -> None:
    """
    Set up a root function for the IDASolver.

    The IDASolver requires two keyword arguments to be set when using root
    functions. The first is 'rootfn' which requires a Callable. The second
    is 'num_events' with allocates memory to an array that stores the root
    function values.

    Parameters
    ----------
    limits : tuple[str, float]
        A tuple of root function criteria arranged as ordered pairs of limit
        names and values, e.g., ('time_h', 10., 'voltage_V', 4.2).
    kwargs : dict
        The IDASolver keyword argumnents dictionary. Both the 'eventsfn' and
        'num_events' keyword arguments must be added to 'kwargs'.

    Returns
    -------
    None.

    """

    rootfn = _RootFunction(limits)

    kwargs['eventsfn'] = rootfn
    kwargs['num_events'] = rootfn.size


def _yaml_reader(file: str) -> dict:
    """
    Read a yaml file.

    This yaml reader has a custom ``!eval`` constructor, which allows you to
    evaluate ``np`` expressions and ``lambda`` functions.

    Parameters
    ----------
    file : str
        An absolute or relative path to a '.yaml' file. The extension will
        be added if not included.

    Raises
    ------
    ValueError
        Invalid file. Only supports '.yaml' files.
    FileNotFoundError
        File does not exist.

    Returns
    -------
    data : dict
        Data dictionary corresponding to the input file.

    """

    _, extension = os.path.splitext(file)

    if extension == '':
        file += '.yaml'
    elif extension != '.yaml':
        raise ValueError("Invalid file. Only supports '.yaml' files.")

    here = os.path.dirname(__file__)
    templates = here + '/templates'

    if file in os.listdir(templates):
        short_warn(f"Using the default parameter file '{file}'.")
        file = templates + '/' + file

    def eval_constructor(loader, node):
        return eval(node.value)

    if not os.path.exists(file):
        raise FileNotFoundError(f"File '{file}' does not exist.")

    reader = YAML(typ='safe')

    add_constructor('!eval', eval_constructor, constructor=SafeConstructor)

    with open(file, 'r') as f:
        data = reader.load(f)

    return data


def formatwarning(message, category, filename, lineno, line=None):
    """Shortened warning format - used for parameter/pre warnings."""
    return f"\n[thevenin {category.__name__}] {message}\n\n"


def short_warn(message, category=None, stacklevel=1, source=None):
    """Print a warning with the short format from ``formatwarning``."""
    original_format = warnings.formatwarning

    warnings.formatwarning = formatwarning
    warnings.warn(message, category, stacklevel, source)

    warnings.formatwarning = original_format



from __future__ import annotations

import numpy as np


class StepFunction:
    """Piecewise step function."""

    __slots__ = ('_tp', '_yp', '_func',)

    def __init__(self, tp: np.ndarray, yp: np.ndarray, y0: float = 0.,
                 ignore_nan: bool = False) -> None:
        """
        Construct a piecewise step function given the times at which step
        changes occur and the values for each time interval. For example,

        .. code-block:: python

            tp = np.array([0, 5])
            yp = np.array([-1, 1])

            y = StepFunction(tp, yp, np.nan)

        Corresponds to

        .. code-block:: python

            if t < 0:
                y = np.nan
            elif 0 <= t < 5:
                y = -1
            else:
                y = 1

        Parameters
        ----------
        tp : 1D np.array
            Times at which a step change occurs [s].
        yp : 1D np.array
            Constant values for each time interval.
        y0 : float, optional
            Value to return when t < tp[0]. In addition to standard float
            values, np.nan and np.inf are supported. The default is 0.
        ignore_nan : bool, optional
            Whether or not to ignore NaN inputs. For NaN inputs, the callable
            returns NaN when False (default) or yp[-1] when True.

        Raises
        ------
        ValueError
            tp and yp must both be 1D.
        ValueError
            tp and yp must be same size.
        ValueError
            tp must be strictly increasing.

        Examples
        --------
        >>> tp = np.array([0, 1, 5])
        >>> yp = np.array([-1, 0, 1])
        >>> func = StepFunction(tp, yp, np.nan)
        >>> print(func(np.array([-10, 0.5, 4, 10])))
        [nan  -1.  0.  1.]

        """

        if tp.ndim != 1 or yp.ndim != 1:
            raise ValueError("tp and yp must both be 1D.")

        if tp.size != yp.size:
            raise ValueError("tp and yp must be same size.")

        if any(np.diff(tp) <= 0.):
            raise ValueError("tp must be strictly increasing.")

        self._tp = tp
        self._yp = np.concatenate(([y0], yp, [yp[-1]]))

        def func(t):

            t = np.asarray(t)
            if t.size == 1:
                t = np.atleast_1d(t)

            idx = np.searchsorted(tp, t, side='right')
            idx = np.clip(idx, 0, tp.size)

            y = self._yp[idx]

            if not ignore_nan:
                y[np.isnan(t)] = np.nan

            if y.size == 1:
                return y.item()
            else:
                return y

        self._func = func

    def __repr__(self) -> str:  # pragma: no cover
        return f"StepFunction(num_steps={self._tp.size})"

    def __call__(self, t: np.ndarray) -> np.ndarray:
        return self._func(t)


class RampedSteps:
    """Step function with ramps."""

    __slots__ = ('_tp', '_yp', '_func', '_t_ramp',)

    def __init__(self, tp: np.ndarray, yp: np.ndarray, t_ramp: float,
                 y0: float = 0.) -> None:
        """
        This class acts like StepFunction, with the same tp, yp, and y0, but
        step transitions include ramps with duration t_ramp. Generally, this
        profile will be more stable compared to a StepFunction profile.

        Parameters
        ----------
        tp : 1D np.array
            Times at which a step change occurs [seconds].
        yp : 1D np.array
            Constant values for each time interval.
        t_ramp : float
            Ramping time between step transitions [seconds].
        y0 : float
            Value to return when t < tp[0]. In addition to standard float
            values, np.nan and np.inf are supported. The default is 0.

        Raises
        ------
        ValueError
            tp and yp must both be 1D.
        ValueError
            tp and yp must be same size.
        ValueError
            t_ramp must be strictly positive.
        ValueError
            tp must be strictly increasing.

        See also
        --------
        StepFunction :
            Uses hard discontinuous steps rather than ramped steps. Generally
            non-ideal for simulations, but may be useful elsewhere.

        """

        if tp.ndim != 1 or yp.ndim != 1:
            raise ValueError("tp and yp must both be 1D.")

        if tp.size != yp.size:
            raise ValueError("tp and yp must be same size.")

        if t_ramp <= 0.:
            raise ValueError("t_ramp must be strictly positive.")

        if any(np.diff(tp) <= 0.):
            raise ValueError("tp must be strictly increasing.")

        tp = np.concatenate((tp, tp + t_ramp))
        yp = np.concatenate(([y0], yp[:-1], yp))

        argsort = np.argsort(tp)

        self._tp = tp[argsort]
        self._yp = yp[argsort]
        self._t_ramp = t_ramp

        self._func = lambda t: np.interp(t, self._tp, self._yp)

    def __repr__(self) -> str:  # pragma: no cover

        num_steps = self._tp.size
        t_ramp = self._t_ramp

        return f"RampedSteps(num_steps={num_steps}, t_ramp={t_ramp:.2e})"

    def __call__(self, t: float) -> float:
        return self._func(t)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from numpy import ndarray


def get_colors(size: int, data: ndarray = None, norm: ndarray = None,
               alpha: float = 1., cmap: str = 'jet'):
    """
    Sample colors from 'cmap'.

    Return a list of colors from a specified colormap. Default options will
    provide evenly spaced colors across 'cmap'. Provide 'data' and/or 'norm'
    to control the ordering, spacing, and normalization.

    Parameters
    ----------
    size : int
        Number of colors to return.
    data : array_like[float] or None, optional
        A 1D array with length 'size' that controls the spacing and sorting of
        the output. By default, spacing is equal and sorting matches 'cmap'.
    norm : array_like[float] or None, optional
        An array-like (min, max) pair that normalizes the colormap to 'data'.
        By default (0, size) if 'data=None' or min/max of 'data' otherwise.
    alpha : float, optional
        Transparency to apply over the colormap. Must be in the range [0, 1].
        The default is 1.
    cmap : str, optional
        A valid matplotlib colormap name. The default is 'jet'.

    Returns
    -------
    colors : list
        A list of (r, g, b, a) color codes.

    Raises
    ------
    ValueError
        'data' length must match 'size'.
    ValueError
        'norm' length must equal 2.

    """

    import numpy as np
    import matplotlib as mpl

    if data is None:
        data = np.arange(size)
    elif len(data) != size:
        raise ValueError(f"{len(data)=} does not match {size=}.")

    if norm is None:
        norm = (min(data), max(data))
    elif len(norm) != 2:
        raise ValueError(f"{len(norm)=} does not match 2.")

    cmap = mpl.colormaps[cmap]

    norm = mpl.colors.Normalize(vmin=norm[0], vmax=norm[1])
    sm = mpl.pyplot.cm.ScalarMappable(cmap=cmap, norm=norm)

    colors = [sm.to_rgba(x, alpha=alpha) for x in data]

    return colors



from __future__ import annotations
from typing import Callable

import numpy as np


class Experiment:
    """Experiment builder."""

    __slots__ = ('_steps', '_kwargs', '_options',)

    def __init__(self, **kwargs) -> None:
        """
        A class to define an experimental protocol. Use the add_step() method
        to add a series of sequential steps. Each step defines a control mode,
        a constant or time-dependent load profile, a time span, and optional
        limiting criteria to stop the step early if a specified event/state is
        detected.

        Parameters
        ----------
        **kwargs : dict, optional
            IDASolver keyword arguments that span all steps.

        See also
        --------
        ~thevenin.IDASolver :
            The solver class, with documentation for most keyword arguments
            that you might want to adjust.

        """

        self._steps = []
        self._kwargs = []
        self._options = kwargs.copy()

    def __repr__(self) -> str:  # pragma: no cover
        """
        Return a readable repr string.

        Returns
        -------
        readable : str
            A console-readable instance representation.

        """

        keys = ['num_steps', 'options']
        values = [self.num_steps, self._options]

        summary = "\n    ".join(f"{k}={v}," for k, v in zip(keys, values))

        readable = f"Experiment(\n    {summary}\n)"

        return readable

    @property
    def steps(self) -> list[dict]:
        """
        Return steps list.

        Returns
        -------
        steps : list[dict]
            List of the step dictionaries.

        """
        return self._steps

    @property
    def num_steps(self) -> int:
        """
        Return number of steps.

        Returns
        -------
        num_steps : int
            Number of steps.

        """
        return len(self._steps)

    def print_steps(self) -> None:
        """
        Prints a formatted/readable list of steps.

        Returns
        -------
        None.

        """
        with np.printoptions(threshold=6, edgeitems=2):
            for i, step in enumerate(self.steps):
                print(f"\nStep {i}\n" + "-"*20)
                for key, value in step.items():
                    print(f"{key:<7} : {value!r}")

                print(f"options : {self._kwargs[i]!r}")

    def add_step(self, mode: str, value: float | Callable, tspan: tuple,
                 limits: tuple[str, float] = None, **kwargs) -> None:
        """
        Add a step to the experiment.

        Parameters
        ----------
        mode : str
            Control mode, {'current_A', 'current_C', 'voltage_V', 'power_W'}.
        value : float | Callable
            Value of boundary contion mode, in the appropriate units.
        tspan : tuple | 1D np.array
            Relative times for recording solution [s]. Providing a tuple as
            (t_max: float, Nt: int) or (t_max: float, dt: float) constructs
            tspan using ``np.linspace`` or ``np.arange``, respectively. Given
            an array uses the values supplied as the evaluation times. Arrays
            must be monotonically increasing and start with zero. See the notes
            for more information.
        limits : tuple[str, float], optional
            Stopping criteria for the new step, must be entered in sequential
            name/value pairs. Allowable names are {'soc', 'temperature_K',
            'current_A', 'current_C', 'voltage_V', 'power_W', 'capacity_Ah',
            'time_s', 'time_min', 'time_h'}. Values for each limit should
            immediately follow a corresponding name and match its units. Time
            limits are in reference to total experiment time. The default is
            None.
        **kwargs : dict, optional
            IDASolver keyword arguments specific to the new step only.

        Returns
        -------
        None.

        Raises
        ------
        ValueError
            'mode' is invalid.
        ValueError
            A 'limits' name is invalid.
        ValueError
            'tspan' tuple must be length 2.
        TypeError
            'tspan[1]' must be type int or float.
        ValueError
            'tspan' arrays must be one-dimensional.
        ValueError
            'tspan[0]' must be zero when given an array.
        ValueError
            'tspan' array length must be at least two.
        ValueError
            'tspan' arrays must be monotonically increasing.

        See also
        --------
        ~thevenin.IDASolver :
            The solver class, with documentation for most keyword arguments
            that you might want to adjust.

        Notes
        -----
        For time-dependent loads, use a Callable for 'value' with a function
        signature like ``def load(t: float) -> float``, where 't' is the step's
        relative time, in seconds.

        Solution times are constructed and saved depending on the 'tspan' input
        types that were supplied:

        * Given (float, int):
            ``tspan = np.linspace(0., tspan[0], tspan[1])``
        * Given (float, float):
            ``tspan = np.arange(0., tspan[0], tspan[1])``

            In this case, 't_max' is also appended to the end. This results
            in the final 'dt' being different from the others if 't_max' is
            not evenly divisible by the given 'dt'.
        * Given 1D np.array:
            When you provide a numpy array it is checked for compatibility.
            If the array is not 1D, is not monotonically increasing, or starts
            with a value other than zero then an error is raised.

        """

        _check_mode(mode)
        _check_limits(limits)

        mode, units = mode.split('_')

        if isinstance(tspan, tuple):

            if not len(tspan) == 2:
                raise ValueError("'tspan' tuple must be length 2.")

            if isinstance(tspan[1], int):
                t_max, Nt = tspan
                tspan = np.linspace(0., t_max, Nt)
            elif isinstance(tspan[1], float):
                t_max, dt = tspan
                tspan = np.arange(0., t_max, dt, dtype=float)
            else:
                raise TypeError("'tspan[1]' must be type int or float.")

            if tspan[-1] != t_max:
                tspan = np.hstack([tspan, t_max])

        else:
            tspan = np.asarray(tspan)

        if tspan.ndim != 1:
            raise ValueError("'tspan' must be one-dimensional.")
        elif tspan[0] != 0.:
            raise ValueError("'tspan[0]' must be zero.")
        elif tspan.size < 2:
            raise ValueError("'tspan' array length must be at least two.")
        elif not all(np.diff(tspan) > 0.):
            raise ValueError("'tspan' must be monotonically increasing.")

        step = {}
        step['mode'] = mode
        step['value'] = value
        step['units'] = units
        step['tspan'] = tspan
        step['limits'] = limits

        self._steps.append(step)
        self._kwargs.append({**self._options, **kwargs})


def _check_mode(mode: str) -> None:
    """
    Check the operating mode.

    Parameters
    ----------
    mode : str
        Operating mode and units.

    Returns
    -------
    None.

    Raises
    ------
    ValueError
        'mode' is invalid.

    """

    valid = ['current_A', 'current_C', 'voltage_V', 'power_W']

    if mode not in valid:
        raise ValueError(f"{mode=} is invalid; valid values are {valid}.")


def _check_limits(limits: tuple[str, float]) -> None:
    """
    Check the limit criteria.

    Parameters
    ----------
    limit : tuple[str, float]
        Stopping criteria and limiting value.

    Returns
    -------
    None.

    Raises
    ------
    ValueError
        'limits' length must be even.
    ValueError
        A 'limits' name is invalid.

    """

    valid = ['soc', 'temperature_K', 'current_A', 'current_C', 'voltage_V',
             'power_W', 'capacity_Ah', 'time_s', 'time_min', 'time_h']

    if limits is None:
        pass
    elif len(limits) % 2 != 0:
        raise ValueError("'limits' length must be even.")
    else:

        for i in range(len(limits) // 2):
            name = limits[2*i]
            value = limits[2*i + 1]

            if name not in valid:
                raise ValueError(f"The limit name '{name}' is invalid; valid"
                                 f" values are {valid}.")

            elif not isinstance(value, (int, float)):
                raise TypeError(f"Limit '{name}' value must be type float.")



"""
Summary
=======
The Thevenin equivalent circuit model is a common low-fidelity battery model
consisting of a single resistor in series with any number of RC pairs, i.e.,
parallel resistor-capacitor pairs. This Python package contains an API for
building and running experiments using Thevenin models. When referring to the
model itself, we use capitalized "Thevenin", and for the package lowercase
``thevenin``.

Accessing the Documentation
---------------------------
Documentation is accessible via Python's ``help()`` function which prints
docstrings from a package, module, function, class, etc. You can also access
the documentation by visiting the website, hosted on Read the Docs. The website
includes search functionality and more detailed examples.

"""

# core package
from ._ida_solver import IDASolver, IDAResult
from ._experiment import Experiment
from ._model import Model
from ._solutions import StepSolution, CycleSolution

# submodules
from . import loadfns
from . import plotutils

__version__ = '0.1.3'

__all__ = [
    'IDASolver',
    'IDAResult',
    'Experiment',
    'Model',
    'StepSolution',
    'CycleSolution',
    'loadfns',
    'plotutils',
]



from __future__ import annotations
from typing import Iterable, TYPE_CHECKING

import textwrap
from copy import deepcopy

import atexit
import numpy as np
import matplotlib.pyplot as plt

from scipy.integrate import cumulative_trapezoid

from ._ida_solver import IDAResult

if TYPE_CHECKING:  # pragma: no cover
    from ._model import Model


class BaseSolution(IDAResult):
    """Base solution."""

    def __init__(self) -> None:
        """
        The base solution class is a parent class to both the StepSolution
        and CycleSolution classes. Inheriting from this class gives each
        solution instance a 'vars' dictionary, access to the 'plot' method,
        and ensures that the slicing of the solution vector into 'vars' is
        consistent between all solutions.

        """

        self.vars = {}

    def __repr__(self) -> str:  # pragma: no cover
        """
        Return a readable repr string.

        Returns
        -------
        readable : str
            A console-readable instance representation.

        """

        classname = self.__class__.__name__

        def wrap_string(label: str, value: list, width: int):
            if isinstance(value, Iterable):
                value = list(value)
            else:
                value = [value]

            indent = ' '*(len(label) + 1)

            if classname == 'StepSolution' and len(value) == 1:
                text = label + f"{value[0]!r}"
            else:
                text = label + "[" + ", ".join(f"{v!r}" for v in value) + "]"

            return textwrap.fill(text, width=width, subsequent_indent=indent)

        data = [
            wrap_string('    success=', self.success, 79),
            wrap_string('    status=', self.status, 79),
            wrap_string('    nfev=', self.nfev, 79),
            wrap_string('    njev=', self.njev, 79),
            wrap_string('    vars=', self.vars.keys(), 79),
        ]

        summary = f"    solvetime={self.solvetime},"
        for d in data:
            summary += f"\n{d},"

        readable = f"{classname}(\n{summary}\n)"

        return readable

    def plot(self, x: str, y: str, show_plot: bool = True, **kwargs) -> None:
        """
        Plot any two variables in 'vars' against each other.

        Parameters
        ----------
        x : str
            A variable key in 'vars' to be used for the x-axis.
        y : str
            A variable key in 'vars' to be used for the y-axis.
        show_plot : bool, optional
            For non-interactive environments only. When True (default) this
            registers `plt.show()` to run at the end of the program. If False,
            you must call `plt.show()` manually.
        **kwargs : dict, optional
            Keyword arguments to pass through to `plt.plot()`. For more info
            please refer to documentation for `maplotlib.pyplot.plot()`.

        Returns
        -------
        None.

        """

        plt.figure()
        plt.plot(self.vars[x], self.vars[y], **kwargs)

        if '_' in x:
            variable, units = x.split('_')
            xlabel = variable.capitalize() + ' [' + units + ']'
        else:
            xlabel = x

        if '_' in y:
            variable, units = y.split('_')
            ylabel = variable.capitalize() + ' [' + units + ']'
        else:
            ylabel = y

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        if show_plot and not plt.isinteractive():
            atexit.register(plt.show)

    def _to_dict(self) -> None:
        """
        Fills the 'vars' dictionary by slicing the SolverReturn solution
        states. Users should generally only access the solution via 'vars'
        since names are more intuitive than interpreting 'y' directly.

        Returns
        -------
        None.

        """

        model = self._model
        ptr = model._ptr

        time = self.t

        soc = self.y[:, ptr['soc']]
        T_cell = self.y[:, ptr['T_cell']]*model.T_inf
        eta_j = self.y[:, ptr['eta_j']]
        voltage = self.y[:, ptr['V_cell']]

        try:
            ocv = model.ocv(soc)
            R0 = model.R0(soc, T_cell)

            assert isinstance(ocv, np.ndarray)
            assert isinstance(R0, np.ndarray)
            assert ocv.shape == soc.shape
            assert R0.shape == soc.shape

        except (TypeError, AssertionError):

            ocv = np.empty_like(soc)
            R0 = np.empty_like(soc)
            for i in range(soc.size):
                ocv[i] = model.ocv(soc[i])
                R0[i] = model.R0(soc[i], T_cell[i])

        current = -(voltage - ocv + np.sum(eta_j, axis=1)) / R0
        capacity = cumulative_trapezoid(-current, x=time/3600., initial=0.)

        # stored time
        self.vars['time_s'] = time
        self.vars['time_min'] = time / 60.
        self.vars['time_h'] = time / 3600.

        # from state variables
        self.vars['soc'] = soc
        self.vars['temperature_K'] = T_cell
        self.vars['voltage_V'] = voltage

        # post-processed variables
        self.vars['current_A'] = current
        self.vars['power_W'] = current*voltage
        self.vars['capacity_Ah'] = soc[0]*model.capacity + capacity
        self.vars['eta0_V'] = current*R0

        for j, eta in enumerate(eta_j.T, start=1):
            self.vars['eta' + str(j) + '_V'] = eta


class StepSolution(BaseSolution):
    """Single-step solution."""

    def __init__(self, model: Model, ida_soln: IDAResult,
                 timer: float) -> None:
        """
        A solution instance for a single experimental step.

        Parameters
        ----------
        model : Model
            The model instance that was run to produce the solution.
        ida_soln : IDAResult
            The unformatted solution returned by IDASolver.
        timer : float
            Amount of time it took for IDASolver to perform the integration.

        """

        super().__init__()

        self._model = deepcopy(model)

        self.message = ida_soln.message
        self.success = ida_soln.success
        self.status = ida_soln.status

        self.t = ida_soln.t
        self.y = ida_soln.y
        self.yp = ida_soln.yp

        self.i_events = ida_soln.i_events
        self.t_events = ida_soln.t_events
        self.y_events = ida_soln.y_events
        self.yp_events = ida_soln.yp_events

        self.nfev = ida_soln.nfev
        self.njev = ida_soln.njev

        self._timer = timer

        self._to_dict()

    @property
    def solvetime(self) -> str:
        """
        Print a statement specifying how long IDASolver spent integrating.

        Returns
        -------
        solvetime : str
            An f-string with the solver integration time in seconds.

        """
        return f"{self._timer:.3f} s"


class CycleSolution(BaseSolution):
    """All-step solution."""

    def __init__(self, *soln: StepSolution, t_shift: float = 1e-3) -> None:
        """
        A solution instance with all experiment steps stitch together into
        a single cycle.

        Parameters
        ----------
        *soln : StepSolution
            All unpacked StepSolution instances to stitch together. The given
            steps should be given in the same sequential order that they were
            run.
        t_shift : float
            Time (in seconds) to shift step solutions by when stitching them
            together. If zero the end time of each step overlaps the starting
            time of its following step. The default is 1e-3.

        """

        super().__init__()

        self._solns = soln
        self._model = soln[0]._model

        sv_size = self._model._sv0.size

        self.message = []
        self.success = []
        self.status = []

        self.t = np.empty([0])
        self.y = np.empty([0, sv_size])
        self.yp = np.empty([0, sv_size])

        self.t_events = None
        self.y_events = None
        self.yp_events = None

        self.nfev = []
        self.njev = []

        self._timers = []

        for soln in self._solns:
            if self.t.size > 0:
                shift_t = self.t[-1] + soln.t + t_shift
            else:
                shift_t = soln.t

            if soln.t_events and self.t.size > 0:
                shift_t_events = self.t[-1] + soln.t_events + t_shift
            elif soln.t_events:
                shift_t_events = soln.t_events

            self.message.append(soln.message)
            self.success.append(soln.success)
            self.status.append(soln.status)

            self.t = np.hstack([self.t, shift_t])
            self.y = np.vstack([self.y, soln.y])
            self.yp = np.vstack([self.yp, soln.yp])

            if soln.t_events:
                if self.t_events is None:
                    self.t_events = shift_t_events
                    self.y_events = soln.y_events
                    self.yp_events = soln.yp_events
                else:
                    self.t_events = np.hstack([self.t_events, shift_t_events])
                    self.y_events = np.vstack([soln.y_events])
                    self.yp_events = np.vstack([soln.yp_events])

            self.nfev.append(soln.nfev)
            self.njev.append(soln.njev)

            self._timers.append(soln._timer)

        self._to_dict()

    @property
    def solvetime(self) -> str:
        """
        Print a statement specifying how long IDASolver spent integrating.

        Returns
        -------
        solvetime : str
            An f-string with the total solver integration time in seconds.

        """
        return f"{sum(self._timers):.3f} s"

    def get_steps(self, idx: int | tuple) -> StepSolution | CycleSolution:
        """
        Return a subset of the solution.

        Parameters
        ----------
        idx : int | tuple
            The step index (int) or first/last indices (tuple) to return.

        Returns
        -------
        :class:`StepSolution` | :class:`CycleSolution`
            The returned solution subset. A StepSolution is returned if 'idx'
            is an int, and a CycleSolution will be returned for the range of
            requested steps when 'idx' is a tuple.

        """

        if isinstance(idx, int):
            return deepcopy(self._solns[idx])
        elif isinstance(idx, (tuple, list)):
            solns = self._solns[idx[0]:idx[1] + 1]
            return CycleSolution(*solns)



from __future__ import annotations

import numpy as np


class Ramp:
    """Linearly ramping load."""

    __slots__ = ('_m', '_b',)

    def __init__(self, m: float, b: float = 0.) -> None:
        """
        A load profile that continuously ramps with slope m.

        Parameters
        ----------
        m : float
            Slope [units/s].
        b : float, optional
            Y-intercept [units]. The default is 0.

        """
        self._m = m
        self._b = b

    def __repr__(self) -> str:  # pragma: no cover
        return f"Ramp(m={self._m:.2e}, b={self._b:.2e})"

    def __call__(self, t: float) -> float:
        return self._m*t + self._b


class Ramp2Constant:
    """Ramp to a constant load."""

    __slots__ = ('_m', '_b', '_step', '_sharpness',)

    def __init__(self, m: float, step: float, b: float = 0.,
                 sharpness: float = 100.) -> None:
        """
        A load profile that ramps with slope m unil the constant step value
        is reached, after which, the load is equal to the step constant. A
        sigmoid is used to smooth the transition between the two piecewise
        functions. Use a large 'sharpness' to reduce smoothing effects.

        Parameters
        ----------
        m : float
            Slope [units/s].
        step : float
            Constant step value [units].
        b : float, optional
            Y-intercept [units]. The default is 0.
        sharpness : float, optional
            How sharp to make the transition between the ramp and step. Low
            values will smooth the transition more. The default is 100.

        Raises
        ------
        ValueError
            m = 0. and m = inf are invalid slopes.
        ValueError
            Cannot reach step with m > 0. and b >= step.
        ValueError
            Cannot reach step with m < 0. and b <= step.
        ValueError
            'sharpness' must be strictly positive.

        """

        if m == 0. or abs(m) == np.inf:
            raise ValueError("m = 0. and m = inf are invalid slopes.")
        elif m > 0. and b >= step:
            raise ValueError("Cannot reach step with m > 0. and b >= step.")
        elif m < 0. and b <= step:
            raise ValueError("Cannot reach step with m < 0. and b <= step.")

        if sharpness <= 0:
            raise ValueError("'sharpness' must be strictly positive.")

        self._m = m
        self._b = b
        self._step = step
        self._sharpness = np.sign(m)*sharpness

    def __repr__(self) -> str:  # pragma: no cover

        data = {'m': self._m, 'b': self._b, 'step': self._step}

        summary = ", ".join([f"{k}={v:.2e}" for k, v in data.items()])

        return f"Ramp2Constant({summary})"

    def __call__(self, t: float) -> float:

        linear = self._m*t + self._b

        z = self._sharpness*(linear - self._step)
        sigmoid = 1. / (1. + np.exp(-np.clip(z, -700, None)))

        return (1. - sigmoid)*linear + sigmoid*self._step



"""
Plotting Utilities
------------------
A module designed to enhance plotting with the matplotlib library. Helper
functions include routines for simplifying color scheme management, formatting
axis ticks, fonts, and more, making it easier to create polished and consistent
visualizations.

"""

from ._colors import get_colors

__all__ = ['get_colors']



"""
Load Functions
--------------
This module contains classes to help construct time-varying load profiles.
All of the classes are callable after construction and take in a value of
time in seconds. Most load functions include a linear ramp that "smooths"
transitions from rest to a constant load, or between constant steps. Using
ramps helps the solver maintain stability when a boundary condition sharply
changes from one value to another, e.g., jumping from rest into a high-rate
charge or discharge. For example, in some cases the solver may crash for a
high-rate discharge.

"""

from ._ramps import Ramp, Ramp2Constant
from ._steps import StepFunction, RampedSteps

__all__ = [
    'Ramp',
    'Ramp2Constant',
    'StepFunction',
    'RampedSteps',
]
